﻿Type=Class
Version=3
@EndOfDesignText@
' clsSQLiteBackup - Class module

' Version: 	1.0
' Date:		02/02/2014
' Author:	LucaMs


' Depends on: KeyValueStore class (which in turn depends on RandomAccessFile library).

' Note: you should declare (and initialize) a global variable named KVStore as KeyValueStore
' in the Process_Globals of Main activity (or elsewhere, but in this case, you should modify
' the references in these sub routines: Initialize, LoadBackupMap and SaveBackupMap).

Sub Class_Globals

	Type typBkpInfo(Created As Long, Note As String)

	Private mBkpExtension As String = "dbbkp"

	Private mAppDir As String				' Application folder name
	
	Private mDB As SQL						' the Database

	Private mDBDir As String				' DB folder name
	Private mDBName As String				' DB name

	Private mBkpDir As String				' backup file folder
	Private mBkpInfos As Map				' a Map of typBkpInfo - BkpName as key
	
	Private mMaxBackups As Int				' Max number of backups
	
	Private mAuto As Boolean				' Periodic Backup ON/OFF
	Private mAutoPeriod As Int				' Period type: DAILY - WEEKLY - MONTHLY
	
End Sub

'
' You should use this class almost as if it were a Singleton class: one instance only.
'
' AppDir - Application folder: will contain the map of backup names;
'
' DB - the SQL Database;
'
' BkpDir - Is the backup folder name: if you pass an empty string ("") then the Root folder will be used;
'
' CreateBkpDir - if set to True, the BkpDir folder will be created, if it does not already exists.
Public Sub Initialize(AppDir As String, _
							 DB As SQL, _
							 DBDir As String, DBName As String, _
							 BkpDir As String, CreateBkpDir As Boolean, _
							 MaxBackups  As Int) _
			  As Boolean

	Dim Res As Boolean = True
	
	If BkpDir = "" Then
		mBkpDir = File.DirRootExternal
	Else
		If Not(File.Exists(BkpDir, "")) Then
			If CreateBkpDir Then
				' TODO check parent
				File.MakeDir("", BkpDir)
			Else
				Log(BkpDir & " don't exists.")
				Res = False
			End If
		End If
	End If
	
	If Res Then
		mAppDir = AppDir
		mDB = DB
		mDBDir = DBDir
		mDBName = DBName
		mBkpDir = BkpDir
		mMaxBackups = MaxBackups
		mBkpInfos = Main.KVStore.GetObject("BkpInfos")
		ReadSettings
		InitBkpInfosMap
		CheckBackupFilesExist
		If mAuto Then
			ManageAuto
		End If
	End If
	
	Return Res
End Sub


#Region PUBLIC PROPERTIES

#Region  + MaxBackups 
' Sets/Gets max number of backups
Public Sub setMaxBackups(MaxBackups As Int)
	mMaxBackups = MaxBackups
	SaveSettings
End Sub
Public Sub getMaxBackups As Int
	Return mMaxBackups
End Sub
#End Region

#Region  + BkpInfos 
' Gets the BkpInfos map.
' Backup names are the Keys;
' Objects of type typBkpInfo are the values
Public Sub getBkpInfos As Map
	Return mBkpInfos
End Sub
#End Region

#Region  + Auto 
' Sets/Gets the periodically automated setting
Public Sub setAuto(Auto As Boolean)
	mAuto = Auto
	SaveSettings
End Sub
Public Sub getAuto As Boolean
	Return mAuto
End Sub
#End Region

#Region  + AutoPeriod 
' Sets/Gets the periodically automated period.
' Period can be: AUTO_DAILY, AUTO_WEEKLY, AUTO_MONTHY
Public Sub setAutoPeriod(Period As Int)
	mAutoPeriod = Period
	SaveSettings
End Sub
Public Sub getAutoPeriod As Int
	Return mAutoPeriod
End Sub
#End Region


#Region  SIMULATED CONSTANTS 

#Region + AUTO 

Public Sub getAUTO_DAILY As Int
	Return 1
End Sub

Public Sub getAUTO_WEEKLY As Int
	Return 2
End Sub

Public Sub getAUTO_MONTHLY As Int
	Return 3
End Sub

#End Region

#End Region

#End Region


#Region PUBLIC METHODS 

#Region --- IO Backup

' Creates a db backup.
Public Sub CreateBackup(BackupName As String, Note As String) As Boolean 'ignore
	CheckBackupFilesExist
	
	If BackupExists(BackupName) Then
		Log(BackupName & " already exists.")
		Return False
	End If
	
	Try
	
		File.Copy(mDBDir, mDBName, mBkpDir, AddExt(BackupName))
		
		Dim BkpInfo As typBkpInfo
		BkpInfo.Initialize
		BkpInfo.Created = DateTime.Now
		BkpInfo.Note = Note

		' If max num of backups is reached, then deletes the oldest one.
		InitBkpInfosMap ' if it is necessary
		If mBkpInfos.Size = mMaxBackups Then
			Dim key As String = mBkpInfos.GetKeyAt(0)
'			' Search for the oldest backup in the map.
'			Dim OldestBkp As typBkpInfo = mBkpInfos.GetValueAt(0)
'			Dim bkp As typBkpInfo
'			For b = 1 To mBkpInfos.Size - 1
'				bkp = mBkpInfos.GetValueAt(b)
'				If bkp.Created < OldestBkp.Created Then
'					OldestBkp = bkp
'					key = mBkpInfos.GetKeyAt(b)
'				End If
'			Next
			mBkpInfos.Remove(key)
			DeleteBackup(key)
		End If
		mBkpInfos.put(BackupName, BkpInfo)
		
	Catch
		
		' Backup failed.
		Log("Backup failed")
		Return False

	End Try
	
	SaveBackupMap
	
	Return True
End Sub

' Restores the db from a backup.
Public Sub RestoreFromBackup(BackupName As String) As Boolean 'ignore
	Dim Res As Boolean = True

	If Not(File.Exists(mBkpDir, AddExt(BackupName))) Then
		Log("File " & BackupName & " not found")
		Return False
	End If
	
	' Creates a temporary copy of current db file
	' for safety in case the restore fails.
	If File.Exists(mAppDir, "BkpTemp.db") Then
		File.Delete(mAppDir, "BkpTemp.db")
	End If
	Try
		mDB.Close
		File.Copy(mAppDir, mDBName, mBkpDir, "BkpTemp.db")
	Catch
		' It was not possible to create a temporary copy:
		' attempts to reopen the original.
		Try
			mDB.Initialize(mAppDir, mDBName, True)
		Catch
			Log("Reopen original db failed.")
		End Try
		Log("Restore failed.")
		Return False
	End Try

	' Tries to restore the db.
	Try
		File.Delete(mAppDir, mDBName)
		File.Copy(mBkpDir, AddExt(BackupName), mAppDir, mDBName)
		' Opens db again
		mDB.Initialize(mAppDir, mDBName, True)
	Catch
		' Restore failed.
		' Restores the original db from temporary db and opens it again.
		File.Copy(mBkpDir, "BkpTemp.db", mAppDir, mDBName)
		mDB.Initialize(mAppDir, mDBName, True)
	End Try

	Return Res
End Sub

' Deletes a backup file.
Public Sub DeleteBackup(BackupName As String) As Boolean 'ignore
	Dim Res As Boolean = True
	
	CheckBackupFilesExist
	
	If Not(mBkpInfos.ContainsKey(BackupName)) OR _
		Not(File.Exists(mBkpDir, AddExt(BackupName))) Then
			Res = False
	Else
		File.Delete(mBkpDir, AddExt(BackupName))
		mBkpInfos.Remove(BackupName)
	End If
	
	Return Res
End Sub

' Gets the backup map from the KeyValueStore
Private Sub LoadBackupMap 'ignore
	mBkpInfos = Main.KVStore.GetObject("BkpInfos")
	InitBkpInfosMap ' if it is necessary
	CheckBackupFilesExist
End Sub

' Puts the backup map to the KeyValueStore
Public Sub SaveBackupMap 'ignore
	Main.KVStore.PutObject("BkpInfos", mBkpInfos)
End Sub

#End Region

#Region --- IO Settings

' Reads: max number of backups, Automatic state and period
Public Sub ReadSettings

	If Main.KVStore.ContainsKey("MaxBackups") Then
		mMaxBackups = Main.KVStore.GetSimple("MaxBackups")
	Else
		mMaxBackups = 5
		Main.KVStore.PutSimple("MaxBackups", mMaxBackups)
	End If

	' Automatic backup ON/OFF
	If Main.KVStore.ContainsKey("AutoBackupONOFF") Then
		mAuto = Main.KVStore.GetObject("AutoBackupONOFF")
	Else
		mAuto = True
		Main.KVStore.PutObject("AutoBackupONOFF", mAuto)
	End If

	' Automatic backup period
	If Main.KVStore.ContainsKey("AutoBackupPeriod") Then
		mAutoPeriod = Main.KVStore.GetSimple("AutoBackupPeriod")
	Else
		mAutoPeriod = 2 ' weekly
		Main.KVStore.PutSimple("AutoBackupPeriod", mAutoPeriod)
	End If

End Sub

' Saves: max number of backups, Automatic state and period
Public Sub SaveSettings
	Main.KVStore.PutSimple("MaxBackups", mMaxBackups)
	Main.KVStore.PutObject("AutoBackup", mAuto)
	Main.KVStore.PutSimple("AutoBackupPeriod", mAutoPeriod)
End Sub

#End Region

' Fills passed ListView with backups data map
Public Sub FillListView(lvw As ListView)
	CheckBackupFilesExist
	
	Dim BkpInfo As typBkpInfo
	Dim bkpName As String
	Dim DateAndTime As String
	
	LoadBackupMap
	
	lvw.Clear
	If mBkpInfos.IsInitialized Then
		' The most recent backup file is the last in the map
		For b = mBkpInfos.Size - 1 To 0 Step - 1
			bkpName = mBkpInfos.GetKeyAt(b)
			BkpInfo = mBkpInfos.Get(bkpName)
			DateAndTime = DateTime.Date(BkpInfo.Created) & " " & DateTime.Time(BkpInfo.Created)
			lvw.AddTwoLines2(bkpName & CRLF & DateAndTime, BkpInfo.Note, bkpName)
		Next
	End If
	
End Sub

' Checks if the BackupName passed is already present in the map of backups
Public Sub BackupExists(BackupName As String) As Boolean
	If mBkpInfos.IsInitialized Then
		Return mBkpInfos.ContainsKey(BackupName)
	Else
		Return False
	End If
End Sub

#End Region


#Region PRIVATE METHODS 

Private Sub InitBkpInfosMap
	If Not(mBkpInfos.IsInitialized) Then
		mBkpInfos.Initialize
	End If
End Sub

Private Sub AddExt(BackupName As String) As String
	If Not(BackupName.ToLowerCase.Contains("." & mBkpExtension)) Then
		BackupName = BackupName & "." & mBkpExtension
	End If
	Return BackupName
End Sub

Private Sub CheckBackupFilesExist

	Dim key As String
	Dim b As Int = 0

	Do While b < mBkpInfos.Size
		key = mBkpInfos.GetKeyAt(b)
		If Not(File.Exists(mBkpDir, AddExt(key))) Then
			mBkpInfos.Remove(key)
		Else
			b = b + 1
		End If
	Loop

End Sub

Private Sub ManageAuto
	
	Dim Note As String = "Automatic daily backup."
	
	' Looks for the file most recent "Auto" file.
	' The most recent is the last in the map.
	Dim bkpName As String
	Dim FoundAuto As Boolean = False
	For b = mBkpInfos.Size - 1 To 0 Step - 1
		bkpName = mBkpInfos.GetKeyAt(b)
		If bkpName.StartsWith("Auto_") Then
			FoundAuto = True
			Exit
		End If
	Next
	
	Dim CreateNewBackup As Boolean = False
	
	If FoundAuto Then
		' Gets the date from the backupfile name.
		Dim LastAutoDate As Long
		LastAutoDate = bkpName.SubString(5)
		
		' Calculates period between last auto backup and today.
		Dim PeriodBetween As Period
		PeriodBetween = DateUtils.PeriodBetween(LastAutoDate, DateTime.Now)

		' Checks if new auto backup file is needed.
		Select Case mAutoPeriod
			Case 1 ' daily
				CreateNewBackup = (PeriodBetween.Days > 0)
			Case 2 ' weekly
				CreateNewBackup = (PeriodBetween.Days > 6)
				Note = Note.Replace("daily", "weekly")
			Case 3 ' monthly
				CreateNewBackup = (PeriodBetween.Months > 0)
				Note = Note.Replace("daily", "monthly")
		End Select
	
	Else
	
		' No auto backup file found, then create new one.
		CreateNewBackup = True
		
	End If
	
	If CreateNewBackup Then
		bkpName = "Auto_" & DateTime.Now
		CreateBackup(bkpName, Note)
	End If
	
End Sub

#End Region